from katie.rl.replay_memory import ReplayMemory
import pytest


def test_replay_memory_creation():
    default_capacity = 10000
    rm = ReplayMemory()
    assert (default_capacity == rm._capacity)


def test_replay_memory_creation_with_custom_capacity():
    for capacity in range(10, 2000, 100):
        rm = ReplayMemory(capacity=capacity)
        assert (capacity == rm._capacity)


def test_replay_memory_batch_sampling():
    capacity = 15000
    batch_size = 200
    rm = ReplayMemory(capacity=capacity)
    for n in range(0, capacity):
        rm.append_memory(n)
    expected_number_of_iterations = capacity / batch_size
    iteration = 0
    for batch in rm.sample_batch(batch_size):
        iteration += 1
    assert (expected_number_of_iterations == iteration)


def test_sample_batch_negative_batch_size():
    batch_size = -50
    rm = ReplayMemory()
    for n in range(0, rm._capacity):
        rm.append_memory(n)
    expected_number_of_iterations = 0
    iteration = 0
    for batch in rm.sample_batch(batch_size):
        iteration += 1
    assert (expected_number_of_iterations == iteration)


def test_sample_batch_zeo_batch_size():
    batch_size = 0
    rm = ReplayMemory()
    for n in range(0, rm._capacity):
        rm.append_memory(n)
    expected_number_of_iterations = 0
    iteration = 0
    for batch in rm.sample_batch(batch_size):
        iteration += 1
    assert (expected_number_of_iterations == iteration)


buffer_size_percentage_capacity_value = [
    (50, 0, True),
    (5000, 50, True),
    (4999, 50, False),
    (5001, 50, True),
    (10000, -1, True),
    (10000, 101, True),
    (0, -1, True),
    (0, 101, False),
    (9999, 100, False)
]

buffer_size_to_default_capacity_value = [
    (50, False),
    (5000, False),
    (0, False),
    (10000, True),
    (9999, False),
    (10001, True)
]


@pytest.mark.parametrize("buffer_size, percentage, expected_value", buffer_size_percentage_capacity_value)
def test_buffer_fulled_by_percentage_value(buffer_size, percentage, expected_value):
    rm = ReplayMemory()
    for n in range(0, buffer_size):
        rm.append_memory(n)
    assert (expected_value == rm.is_buffer_fulled_by_percentage_value(capacity_percantage=percentage))


@pytest.mark.parametrize("buffer_size, expected_value", buffer_size_to_default_capacity_value)
def test_buffer_fulled_by_percentage_value(buffer_size, expected_value):
    rm = ReplayMemory()
    for n in range(0, buffer_size):
        rm.append_memory(n)
    assert (expected_value == rm.is_buffer_full())




import hashlib
import pickle
import numpy as np
from collections import deque

"""
ReplayMemory class gives an ability to store and give back data in random batches.
Memory has a limited buffer based on the FIFO queue.
"""


class ReplayMemory:
    def __init__(self, capacity=10000):
        self._capacity = capacity
        self._buffer = deque()

    def sample_batch(self, batch_size: int):
        """
        Creates an iterator that returns random batches from the buffer.
        The batch size has to be smaller than the current number of data elements.
        :param batch_size: batch size
        :return: iterator returning random batches
        """
        if batch_size <= 0:
            return
        ofs = 0
        vals = list(self._buffer)
        np.random.shuffle(vals)
        while (ofs + 1) * batch_size <= len(self._buffer):
            yield vals[ofs * batch_size:(ofs + 1) * batch_size]
            ofs += 1

    def append_memory(self, data):
        """
        Adds new data element to the replay memory buffer.
        When the capacity is reached, the oldest element in the
        buffer is removed.
        :param data: data element to add
        :return:
        """
        self._buffer.append(data)
        while len(self._buffer) > self._capacity:
            self._buffer.popleft()

    def get_current_buffer_size(self):
        """
        Return the current number of buffer elements
        :return: number of buffer elements
        """
        return len(self._buffer)

    def get_capacity(self):
        """
        Returns the maximum capacity of the memory
        :return: the memory set capacity
        """
        return self._capacity

    def is_buffer_full(self):
        """
        Tells if buffer reached the maximum value of the capacity.
        :return: True when the number of elements is equal of higher than the capacity.
        False otherwise.
        """
        return len(self._buffer) >= self._capacity

    def is_buffer_fulled_by_percentage_value(self, capacity_percantage=100):
        """
        Tells if buffer reached the given percentage of the capacity.
        The percentage value should be between 0 and 100.
        Negative values treat as 0% and value higher than 100 as 100%.
        :param capacity_percantage: the percentage of value of the buffer that we want to compare the capacity.
        :return: True when the elements in the buffer are equal or higher than
        the percentage value of the memory capacity, False otherwise.
        """
        percentage = capacity_percantage
        if percentage > 100:
            percentage = 100
        elif percentage < 0:
            percentage = 0
        return len(self._buffer) >= self._capacity * (percentage / 100.0)

    def save_memory_buffer(self, file_name: str, hash=False):
        """
        Saves the buffer to pickle file.
        :param file_name: the name of the file. The file name must contain the '.pickle' extension.
        :param hash: determine to whether additionaly write hash in separate file or not
        """
        if file_name.endswith(".pickle"):
            with open(file_name, 'wb') as f:
                pickle.dump(self._buffer, f)
            if hash:
                hash_file = file_name[:-7] + ".SHA256"
                _generate_hash_file(file_name, hash_file)
            return
        raise TypeError("The file name {} does not ends with '.pickle' extension.".format(file_name))

    def load_memory_buffer(self, file_name: str, hash_file=None):
        """
        Loads the buffer from the pickle file. Caution - it will clear the current content of the memory buffer.
        The loaded data will not exceed the memory capacity.
        :param file_name: the name of the file. The file name must contain the '.pickle' extension.
        :param hash_file: the '.SHA256' file that contains the hash of the pickle data. If the hash from the file is not
        same as the hash of the file, it raises PersmissionError.
        :raise BufferError: when hashes are different.
        :raise FileNotFoundError: when any of the files are not found.
        :raise TypeError: when pickle file does not have `.pickle` extension or hash file has '.SHA256' extension.
        """
        if file_name.endswith(".pickle"):
            with open(file_name, 'rb') as f:
                if hash_file is not None:
                    if hash_file.endswith(".SHA256"):
                        with open(hash_file, 'rb') as fh:
                            hash_value = fh.readline()
                        pickle_hash = _get_hash_of_file(file_name)
                        if hash_value != pickle_hash:
                            raise BufferError("Hashes {} and {} are different".format(hash_value, pickle_hash))
                    else:
                        raise TypeError("The file name {} does not ends with '.SHA256' extension.".format(hash_file))
                self._buffer.clear()
                for data in pickle.load(f):
                    self._buffer.append(data)
                    if self.is_buffer_full():
                        return
                return
        raise TypeError("The file name {} does not ends with '.pickle' extension.".format(file_name))


def _get_hash_of_file(file):
    sha256_hash = hashlib.sha256()
    with open(file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest().encode(encoding='utf8')


def _generate_hash_file(pickle_file, hash_file):
    hash = _get_hash_of_file(pickle_file)
    print(hash)
    with open(hash_file, "wb") as f:
        f.write(hash)



import os
import pickle
import pytest
from katie.rl.replay_memory import ReplayMemory

_file_name = "buffer.pickle"
_file_name2 = "buffer2.pickle"
_file_name_no_extension = "buffer"
_file_hash = "buffer.SHA256"
_file_hash2 = "buffer2.SHA256"


def _fill_memory(replay_memory: ReplayMemory, data: str = ""):
    for n in range(0, replay_memory.get_capacity()):
        replay_memory.append_memory(str(n) + data)


def test_save_memory():
    rm = ReplayMemory()
    _fill_memory(rm)
    rm.save_memory_buffer(_file_name)
    assert (os.path.exists(_file_name))
    os.remove(_file_name)


def test_save_memory_wrong_file_format():
    rm = ReplayMemory()
    _fill_memory(rm)
    with pytest.raises(TypeError):
        rm.save_memory_buffer(_file_name_no_extension)


def test_save_memory_hash():
    rm = ReplayMemory()
    _fill_memory(rm)
    rm.save_memory_buffer(_file_name, hash=True)
    assert (os.path.exists(_file_hash))
    os.remove(_file_name)
    os.remove(_file_hash)


def test_load_memory():
    rm = ReplayMemory()
    _fill_memory(rm)
    rm.save_memory_buffer(_file_name)
    assert (os.path.exists(_file_name))

    rm_load = ReplayMemory()
    rm_load.load_memory_buffer(_file_name)

    assert (rm._buffer == rm_load._buffer)
    os.remove(_file_name)


def test_load_memory_no_existing_file():
    rm = ReplayMemory()
    with pytest.raises(FileNotFoundError):
        rm.load_memory_buffer(_file_name)


def test_load_memory_wrong_format_file():
    rm = ReplayMemory()
    with pytest.raises(TypeError):
        rm.load_memory_buffer(_file_name_no_extension)


def test_load_memory_buffer_is_cleaned():
    rm = ReplayMemory()
    _fill_memory(rm)
    rm.save_memory_buffer(_file_name)
    assert (os.path.exists(_file_name))

    rm_load = ReplayMemory()
    _fill_memory(rm_load, data="TEST")
    rm_load.load_memory_buffer(_file_name)

    assert (rm._buffer == rm_load._buffer)
    os.remove(_file_name)


def test_load_memory_do_not_corrupt_buffer_when_fails():
    rm = ReplayMemory()
    _fill_memory(rm, data="TEST")
    buffer_copy = rm._buffer.copy()

    with pytest.raises(TypeError):
        rm.load_memory_buffer(_file_name_no_extension)
    assert (buffer_copy == rm._buffer)

    with pytest.raises(FileNotFoundError):
        rm.load_memory_buffer(_file_name2)
    assert (buffer_copy == rm._buffer)


def test_load_memory_keeps_the_set_capacity():
    rm = ReplayMemory(capacity=1000)
    _fill_memory(rm)
    rm.save_memory_buffer(_file_name)

    rm_2 = ReplayMemory(capacity=500)
    rm_2.load_memory_buffer(_file_name)

    assert (len(rm_2._buffer) == rm_2._capacity)
    os.remove(_file_name)


def test_load_memory_not_fails_when_pickle_content_not_iterable():
    test_pickle = 15000
    with open(_file_name, "wb") as f:
        pickle.dump(test_pickle, f)
    rm = ReplayMemory()
    with pytest.raises(TypeError):
        rm.load_memory_buffer(_file_name)
    os.remove(_file_name)


def test_load_hash_memory():
    rm = ReplayMemory()
    _fill_memory(rm)
    rm.save_memory_buffer(_file_name, hash=True)
    assert (os.path.exists(_file_name))
    assert (os.path.exists(_file_hash))

    rm_load = ReplayMemory()
    rm_load.load_memory_buffer(_file_name, _file_hash)

    assert (rm._buffer == rm_load._buffer)
    os.remove(_file_name)
    os.remove(_file_hash)


def test_load_hash_memory_with_wrong_hash():
    rm = ReplayMemory()
    _fill_memory(rm)
    rm.save_memory_buffer(_file_name, hash=True)
    assert (os.path.exists(_file_name))
    assert (os.path.exists(_file_hash))

    rm2 = ReplayMemory()
    _fill_memory(rm2, data="TEST")
    rm2.save_memory_buffer(_file_name2, hash=True)
    assert (os.path.exists(_file_name2))
    assert (os.path.exists(_file_hash2))

    rm_load = ReplayMemory()
    with pytest.raises(BufferError):
        rm_load.load_memory_buffer(_file_name, _file_hash2)
    with pytest.raises(BufferError):
        rm_load.load_memory_buffer(_file_name2, _file_hash)

    os.remove(_file_name)
    os.remove(_file_hash)
    os.remove(_file_name2)
    os.remove(_file_hash2)


def test_load_hash_memory_no_hash_file():
    rm = ReplayMemory()
    _fill_memory(rm)
    rm.save_memory_buffer(_file_name)
    assert (os.path.exists(_file_name))

    rm_load = ReplayMemory()
    with pytest.raises(FileNotFoundError):
        rm_load.load_memory_buffer(_file_name, _file_hash)

    os.remove(_file_name)


def test_load_hash_memory_hash_file_wrong_format():
    rm = ReplayMemory()
    _fill_memory(rm)
    rm.save_memory_buffer(_file_name)
    assert (os.path.exists(_file_name))

    rm_load = ReplayMemory()
    with pytest.raises(TypeError):
        rm_load.load_memory_buffer(_file_name, _file_name_no_extension)

    os.remove(_file_name)



import torch
import pytest
from katie.rl.softmax_body import SoftmaxBody


def test_creation():
    sb = SoftmaxBody()
    assert (1.0 == sb.temperature)


def test_call():
    sb = SoftmaxBody()
    outputs = torch.tensor([[50, 20, 30, 40]])
    sb_output = sb(outputs=outputs)
    assert (torch.is_tensor(sb_output))
    assert ((1, 1) == sb_output.shape)
    assert ((1, 3) == sb(outputs=outputs, num_samples=3).shape)


def test_call_with_bigger_number_of_samples():
    sb = SoftmaxBody()
    outputs = torch.tensor([[50, 20, 30, 40]])
    sb_output = sb(outputs=outputs, num_samples=10)
    assert (torch.is_tensor(sb_output))
    print(sb_output.shape)
    assert ((1, 4) == sb_output.shape)


def test_call_too_many_dimensions_in_outputs_shape():
    sb = SoftmaxBody()
    outputs = torch.tensor([[[50, 20, 30, 40]]])
    with pytest.raises(Exception):
        sb(outputs=outputs)



import setuptools
import os
import re

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def load_version():
    """ Loads a file content """
    filename = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "katie", "__init__.py"))
    with open(filename, "rt") as version_file:
        init_file = version_file.read()
        version = re.search(r"__version__ = '([0-9a-z.-]+)'", init_file).group(1)
        return version


setuptools.setup(name='ai-katie',
                 version=load_version(),
                 description='Utility things for Data Science and AI',
                 long_description=long_description,
                 author='danpeczek',
                 author_email='danpeczek@gmail.com',
                 url='https://github.com/danpeczek/ai-katie',
                 project_urls={
                     "Bug Tracker": "https://github.com/danpeczek/ai-katie/issues",
                 },
                 install_requires=["torch", "numpy"],
                 classifiers=[
                     'Intended Audience :: Developers',
                     'Intended Audience :: Education',
                     'Intended Audience :: Science/Research',
                     'Operating System :: OS Independent',
                     'Programming Language :: Python :: 3',
                     'Programming Language :: Python :: 3.6',
                     'Programming Language :: Python :: 3.7',
                     'Programming Language :: Python :: 3.8',
                     'Programming Language :: Python :: 3.9',
                     'Topic :: Communications :: Email',
                     'Topic :: Scientific/Engineering :: Artificial Intelligence',
                     'Topic :: Scientific/Engineering :: Information Analysis'
                 ],
                 package_dir={"": "."},
                 packages=setuptools.find_packages(where="."),
                 python_requires=">=3.6"
                 )



from torch.nn.functional import softmax

"""
SoftmaxBody applies softmax function to the output tensor to generate probabilities of choosing particular output.
The sum of the outputs after softmax application is 1.
"""


class SoftmaxBody:

    def __init__(self, temperature=1.0):
        """
        Constructs the SoftmaxBody object
        :param temperature: the parameter upon which all probabilities will be multiplied before softmax application
        """
        self.temperature = temperature

    def __call__(self, outputs, num_samples=1):
        """
        Generates softmax upon the outputs and based on the probability returns samples.
        :param outputs: the outputs tensor that we want to apply softmax. The shape of outputs should be (1, n).
        If it's not - the exception is thrown.
        :param num_samples: the number of samples that should be taken based on the probability. Defaults to 1.
        If number of samples is longer than length of outputs along the last dimensions,
         the number of samples is set to length of the last dimension.
        :return: indexes of the chosen outputs taken by the probability of this output.
        """
        probabilities = softmax(outputs * self.temperature, dim=1)
        outputs_size = outputs.size()
        if len(outputs_size) != 2:
            raise Exception("The outputs shape should be (m, n)! But instead got {}".format(outputs.shape))
        if num_samples >= outputs_size[-1]:
            return probabilities.multinomial(num_samples=outputs_size[-1])
        else:
            return probabilities.multinomial(num_samples=num_samples)
